//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://emule.sf.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "partfile.h"
#include "emule.h"

CPartFile::CPartFile(){
	fullname = 0;
	status = PS_UNKNOWN;
	newdate = false;
	partmetfilename = 0;
}

CPartFile::~CPartFile(){
	if (fullname)
		delete[] fullname;
	if (partmetfilename)
		delete[] partmetfilename;

	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos))
		delete gaplist.GetAt(pos);
}

bool CPartFile::LoadPartFile(char* directory,char* in_filename){
	partmetfilename = strdup(in_filename);
	directory = strdup(directory);
	char* buffer = new char[strlen(directory)+strlen(partmetfilename)+2];
	sprintf(buffer,"%s\\%s",directory,partmetfilename);
	fullname = buffer;
	FILE* file = 0;
	try{
		// readfile data form part.met file
		file = fopen(fullname,"rbS");
		if (!file)
			throw "Failed to open part.met file";
		uint8 version;
		fread(&version,1,1,file);
		if (version != PARTFILE_VERSION)
			throw "Invalid part.met fileversion";
		if (!this->LoadDateFromFile(file))
			throw "Failed to load part.met file-date";
		//date = 0; //debug
		if (!this->LoadHashsetFromFile(file))
			throw "Failed to load part.met file-hashset";
		uint32 tagcount;
		fread(&tagcount,4,1,file);
		for (uint32 j = 0; j != tagcount;j++){
			CTag* newtag = new CTag(file);
			switch(newtag->tag->specialtag){
				case FT_FILENAME:{
					filename = strdup(newtag->tag->stringvalue);
					delete newtag;
					break;
				}
				case FT_FILESIZE:{
					filesize = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_TRANSFERED:{
					transfered = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_PRIORITY:{
					priority = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				case FT_STATUS:{
					paused = newtag->tag->intvalue;
					delete newtag;
					break;
				}
				default:{
					// hmmm why are gaptag no spcialtags (name len = 2) ? very odd
					if ((!newtag->tag->specialtag) && newtag->tag->tagname[0] == FT_GAPSTART){					
						Gap_Struct* gap = new Gap_Struct; 
						gap->start = newtag->tag->intvalue;
						delete newtag;
						j++;
						CTag* newtag = new CTag(file);
						if (newtag->tag->tagname[0] != FT_GAPEND){
							delete newtag;
							throw "Invalid Gap-Tag in part.met file";
						}
						gap->end = newtag->tag->intvalue;
						// gap start = first missing byte but gap ends = first non-missing byte in edonkey
						// but I think its easier to user the real limits so:
						gap->end--;
						gaplist.AddTail(gap);
						delete newtag;
					}
					else
						taglist.Add(newtag);
				}
					
			}

		}
	}
	catch(char* error){
		if (file)
			fclose(file);
		theApp.emuledlg->AddLogLine(false,"ERROR: %s (%s => %s)",error,partmetfilename,filename);
		return false;
	}
	fclose(file);
	if (!filesize){
		status = PS_EMPTY;
		return true;
	}
	// check date of .part file - if its wrong, rehash file
	CFileFind ff;
	char* searchpath = strdup(fullname);
	searchpath[strlen(fullname)-4] = 0;
	bool end = !ff.FindFile(searchpath,0);
	delete[] searchpath;
	try{
		if (!end)
			ff.FindNextFile();
		if (end || ff.IsDirectory())
			throw ".part file not found";
		CTime lwtime;
		ff.GetLastWriteTime(lwtime);
		if (date != mktime(lwtime.GetLocalTm())){
			theApp.emuledlg->AddLogLine(false,"Warning: %s might be corrputed, performing rehash of completed parts",buffer,filename);
			// rehash
			status = PS_WAITINGFORHASH;
			CAddFileThread* addfilethread = (CAddFileThread*) AfxBeginThread(RUNTIME_CLASS(CAddFileThread), THREAD_PRIORITY_NORMAL,0, CREATE_SUSPENDED);
			addfilethread->SetValues(0,directory,ff.GetFileName().GetBuffer(),this);
			addfilethread->ResumeThread();	
		}
	}
	catch(char* error){
		ff.Close();
		theApp.emuledlg->AddLogLine(false,"ERROR: %s (%s => %s)",error,partmetfilename,filename);
		return false;
	}
	ff.Close();
	status = PS_READY;
	return true;
}

bool CPartFile::SafePartFile(){

	FILE* file = 0;
	try{
		//get filedate
		if (newdate){
			CFileFind ff;
			char* searchpath = strdup(fullname);
			searchpath[strlen(fullname)-4] = 0;
			bool end = !ff.FindFile(searchpath,0);
			delete[] searchpath;
			if (!end)
				ff.FindNextFile();
			if (end || ff.IsDirectory())
				throw ".part file not found";
			CTime lwtime;
			ff.GetLastWriteTime(lwtime);
			date = mktime(lwtime.GetLocalTm());
			ff.Close();
		}
		// readfile data form part.met file
		file = fopen(fullname,"wbS");
		if (!file)
			throw "Failed to open part.met file";
		//version
		uint8 version = PARTFILE_VERSION;
		fwrite(&version,1,1,file);
		//date
		fwrite(&date,4,1,file);
		//hash
		fwrite(&filehash,16,1,file);
		uint16 parts = hashlist.GetCount();
		fwrite(&parts,2,1,file);
		for (int x = 0; x != parts; x++)
			fwrite(hashlist[x],16,1,file);
		//tags
		uint32 tagcount = taglist.GetCount()+5+(gaplist.GetCount()*2);
		fwrite(&tagcount,4,1,file);
		CTag* nametag = new CTag(FT_FILENAME,filename);
		nametag->WriteTagToFile(file);
		delete nametag;
		CTag* sizetag = new CTag(FT_FILESIZE,filesize);
		sizetag->WriteTagToFile(file);
		delete sizetag;
		CTag* transtag = new CTag(FT_TRANSFERED,transfered);
		transtag->WriteTagToFile(file);
		delete transtag;
		CTag* statustag = new CTag(FT_STATUS,(paused)? 1:0);
		statustag->WriteTagToFile(file);
		delete statustag;
		CTag* prioritytag = new CTag(FT_PRIORITY,priority);
		prioritytag->WriteTagToFile(file);
		delete prioritytag;
		for (uint32 j = 0; j != taglist.GetCount();j++)
			taglist[j]->WriteTagToFile(file);
		//gaps
		char* namebuffer = new char[10];
		char* number = &namebuffer[1];
		uint16 i_pos = 0;
		for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
			itoa(i_pos,number,10);
			namebuffer[0] = FT_GAPSTART;
			CTag* gapstarttag = new CTag(namebuffer,gaplist.GetAt(pos)->start);
			gapstarttag->WriteTagToFile(file);
			// gap start = first missing byte but gap ends = first non-missing byte in edonkey
			// but I think its easier to user the real limits
			namebuffer[0] = FT_GAPEND;
			CTag* gapendtag = new CTag(namebuffer,(gaplist.GetAt(pos)->end)+1);
			gapendtag->WriteTagToFile(file);
			delete gapstarttag;
			delete gapendtag;
			i_pos++;
		}
		delete[] namebuffer;

		if (ferror(file))
			throw "unexpected write error";
	}	
	catch(char* error){
		if (file)
			fclose(file);
		theApp.emuledlg->AddLogLine(false,"ERROR while saving partfile: %s (%s => %s)",error,partmetfilename,filename);
		return false;
	}
	fclose(file);
	return true;
}

void CPartFile::PartFileHashFinished(CKnownFile* result){
	if (result){
		newdate = true;
		bool errorfound = false;
		for (int i = 0; i != hashlist.GetSize(); i++){
			if (IsComplete(i*PARTSIZE,((i+1)*PARTSIZE)-1)){
				if (!(result->GetPartHash(i) && !memcmp(result->GetPartHash(i),this->GetPartHash(i),16))){
					theApp.emuledlg->AddLogLine(false,"Found corrupted part (%i) in %s",i+1,filename);		
					AddGap(i*PARTSIZE,((((i+1)*PARTSIZE)-1) >= filesize) ? filesize-1 : ((i+1)*PARTSIZE)-1);
					errorfound = true;
				}
			}
		}
		delete result;
		if (!errorfound)
			theApp.emuledlg->AddLogLine(false,"Finished rehashing %s, file seems to be ok",filename);
		status = PS_READY;
		theApp.sharedfiles->SafeAddKFile(this);
	}
	else{
		theApp.emuledlg->AddLogLine(false,"Failed to rehash %s. File might be coruppted or defect.",filename);
		status = PS_ERROR;
	}
}

void CPartFile::AddGap(uint32 start, uint32 end){
	bool startset = false;
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
		Gap_Struct* cur_gap = gaplist.GetAt(pos);
		if (cur_gap->start >= start && cur_gap->end <= end){ // this gap is inside the new gap - delete
			gaplist.RemoveAt(pos);
			gaplist.GetPrev(pos);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->start >= start && cur_gap->start <= end){// a part of this gap is in the new gap - extend limit and delete
			end = cur_gap->end;
			gaplist.RemoveAt(pos);
			gaplist.GetPrev(pos);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->end <= end && cur_gap->end >= start){// a part of this gap is in the new gap - extend limit and delete
			start = cur_gap->start;
			gaplist.RemoveAt(pos);
			gaplist.GetPrev(pos);
			delete cur_gap;
			continue;
		}
		else if (start >= cur_gap->start && end <= cur_gap->end){// new gap is already inside this gap - return
			return;
		}
	}
	Gap_Struct* new_gap = new Gap_Struct;
	new_gap->start = start;
	new_gap->end = end;
	gaplist.AddTail(new_gap);
	theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
	newdate = true;
}

bool CPartFile::IsComplete(uint32 start, uint32 end){
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
		Gap_Struct* cur_gap = gaplist.GetAt(pos);
		if ((cur_gap->start >= start && cur_gap->end <= end)||(cur_gap->start >= start 
			&& cur_gap->start <= end)||(cur_gap->end <= end && cur_gap->end >= start)
			||(start >= cur_gap->start && end <= cur_gap->end)){
				return false;	
			}
	}
	return true;
}

void CPartFile::FillGap(uint32 start, uint32 end){

	theApp.emuledlg->transferwnd.downloadlistctrl.UpdateItem(this);
}

void CPartFile::DrawStatusBar(CDC* dc, RECT* rect){
	CBrush br(RGB(0,0,0));
	RECT gaprect;
	dc->FillRect(rect,&CBrush(RGB(220,220,220)));
	gaprect.top = rect->top + 4;
	gaprect.bottom = rect->bottom;
	gaprect.left = rect->left;
	gaprect.right = rect->right;
	dc->FillRect(&gaprect,&br);
	uint32 allgaps = 0;
	float blockpixel = (float)(rect->right - rect->left)/(filesize/1024);
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;gaplist.GetNext(pos)){
		Gap_Struct* cur_gap = gaplist.GetAt(pos);
		gaprect.right = rect->left + (uint32)(((float)cur_gap->start/1024)*blockpixel);
		gaprect.left = rect->left +  (uint32)(((float)cur_gap->end/1024)*blockpixel);
		dc->FillRect(&gaprect,&CBrush(RGB(255,0,0)));
		allgaps += cur_gap->end - cur_gap->start;
	}
	gaprect.bottom = gaprect.top;
	gaprect.top -= 4;
	gaprect.left = rect->left;
	gaprect.right = (uint32)(rect->left+((filesize-allgaps)/1024)*blockpixel);
	dc->FillRect(&gaprect,&CBrush(RGB(0,150,0)));
}

void CPartFile::WritePartStatus(CFile* file){
	uint16 parts = hashlist.GetCount();
	file->Write(&parts,2);
	uint16 done = 0;
	while (done != parts){
		uint8 towrite = 0;
		for (uint32 i = 0;i != 8;i++){
			if (IsComplete(done*PARTSIZE,((done+1)*PARTSIZE)-1))
				towrite |= (1<<i);
			done++;
			if (done == parts)
				break;
		}
		file->Write(&towrite,1);
	}
}
