//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://emule.sf.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "uploadqueue.h"
#include "packets.h"
#include "emule.h"
#include "knownfile.h"



CUploadQueue::CUploadQueue(CPreferences* in_prefs){
	app_prefs = in_prefs;
	h_timer = SetTimer(0,141,100,*TimerProc);
	if (!h_timer)
		theApp.emuledlg->AddLogLine(false,"Fatal Error: Failed to create Timer");
	estadatarate = 2000;
	datarate = 0;
	dataratems = 0;
	dataratetmp = 0;
	datarateave = 0;
	count = 0;

}

void CUploadQueue::AddUpNextClient(CUploadClient* directadd){
	POSITION toadd = 0;
	uint32	bestscore = 0;
	CUploadClient* newclient;
	// select next client or use given client
	if (!directadd){
		for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
			if (waitinglist.GetAt(pos)->GetScore() > bestscore){
				bestscore = waitinglist.GetAt(pos)->GetScore();
				toadd = pos;
			}
		}
		if (!toadd)
			return;
		newclient = waitinglist.GetAt(toadd);
		waitinglist.RemoveAt(toadd);
		theApp.emuledlg->transferwnd.ShowQueueCount(waitinglist.GetCount());
		
	}
	else
		newclient = directadd;

	// tell the client that we are now ready to upload
	if (!newclient->socket->IsConnected()){
		//if (newclient->HasLowID()){
			//TODO: upload contact to lowid clients via server

		//}
		//else{
			char* test = newclient->GetFullIP();
			delete newclient->socket;
			newclient->socket = new CClientReqSocket(app_prefs,false,(CBaseClient*)newclient);
			newclient->socket->Create();
			newclient->socket->Connect(newclient->GetFullIP(),newclient->GetUserPort());
			newclient->socket->SendHelloPacket();
		//}
	}
	else{
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		newclient->socket->SendPacket(packet,true);
	}
	newclient->SetUpStartTime();
	uploadinglist.AddTail(newclient);
	theApp.emuledlg->transferwnd.uploadlistctrl.AddClient(newclient);
}

void CUploadQueue::Process(){
	count++;
	dataratetmp += dataratems;
	if (count == 50){
		datarate = dataratetmp/5;
		count = 0;
		dataratetmp = 0;
	}
	dataratems = 0;
	if (AcceptNewClient() && waitinglist.GetCount())
		AddUpNextClient();
	if (!uploadinglist.GetCount())
		return;

	int16 clientsrdy = 0;
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;uploadinglist.GetNext(pos)){
		if ((!uploadinglist.GetAt(pos)->socket->IsBusy()) && uploadinglist.GetAt(pos)->cur_block)
			clientsrdy++;
	}
	if (!clientsrdy){
		estadatarate -= 200;
		if (estadatarate < 100)
			estadatarate = 100;
		clientsrdy++;
	}
	else{
		estadatarate += 200;
		if (estadatarate > app_prefs->GetMaxUpload()*102)
			estadatarate = app_prefs->GetMaxUpload()*102;
	}
	uint32 sendperclient = estadatarate/clientsrdy;

	for (pos = uploadinglist.GetHeadPosition();pos != 0;uploadinglist.GetNext(pos)){
		if (uploadinglist.GetAt(pos)->socket->IsDisconnected()){
			RemoveFromUploadQueue((CUploadClient*)uploadinglist.GetAt(pos));
			break;
		}
		dataratems += uploadinglist.GetAt(pos)->SendBlockData(sendperclient);
	}
};

bool CUploadQueue::AcceptNewClient(){
	// check if we can allow a new client to start downloading form us
	if (uploadinglist.GetCount() < MIN_UP_CLIENTS_ALLOWED)
		return true;
	else if (uploadinglist.GetCount() >= MAX_UP_CLIENTS_ALLOWED)
		return false;

	//now the final check
	if (uploadinglist.GetCount() < (datarate/UPLOAD_CLIENT_DATARATE))
		return true;
	//nope
	return false;
}

CUploadQueue::~CUploadQueue(){
	KillTimer(0,141);
}

POSITION CUploadQueue::GetWaitingClient(CUploadClient* client){
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
		if (client == waitinglist.GetAt(pos))
			return pos;
	}
	return 0;
}

POSITION CUploadQueue::GetWaitingClientByIP(uint32 ip){
	for (POSITION pos = waitinglist.GetHeadPosition();pos != 0;waitinglist.GetNext(pos)){
		if (ip == waitinglist.GetAt(pos)->GetIP())
			return pos;
	}
	return 0;
}

POSITION CUploadQueue::GetDownloadingClient(CUploadClient* client){
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;uploadinglist.GetNext(pos)){
		if (client == uploadinglist.GetAt(pos))
			return pos;
	}
	return 0;
}

void CUploadQueue::AddClientToQueue(CUploadClient* client){
	if(POSITION pos = GetWaitingClientByIP(client->GetIP())){
		//if he's already on the list, save his startime and remove him
		CUploadClient* oldclient = waitinglist.GetAt(pos);
		client->SetWaitStartTime(oldclient->GetWaitStartTime());
		waitinglist.RemoveAt(pos);
		delete oldclient;
	}
	if (client->IsDownloading()){
		// he's already downloading and wants probably only another file
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		client->socket->SendPacket(packet,true);
		return;
	}
	if (waitinglist.GetCount() == 0 && AcceptNewClient())
		AddUpNextClient(client);
	else{
		waitinglist.AddTail(client);
		theApp.emuledlg->transferwnd.ShowQueueCount(waitinglist.GetCount());
		if (client->socket->IsConnected())
			client->socket->Disconnect();
	}
}
bool CUploadQueue::RemoveFromUploadQueue(CUploadClient* client){
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;uploadinglist.GetNext(pos)){
		if (client == uploadinglist.GetAt(pos)){
			theApp.emuledlg->transferwnd.uploadlistctrl.RemoveClient(uploadinglist.GetAt(pos));
			uploadinglist.RemoveAt(pos);	
			return true;
		}
	}
	return false;
}

VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg,UINT_PTR idEvent,DWORD dwTime){
	theApp.uploadqueue->Process();
	i++;
	if (i == 50){
		i = 0;
		theApp.emuledlg->ShowTransferRate();
	}	
}